/*
Copyright (C) 2024 Christoph Boelitz

E-mail: tools@dosforum.de
Project page: http://tools.dosforum.de/

This program is free software as long as its used privately.
Commercial use, as unlikely as it may seem, is hereby prohibited.
The source code may be modified and distributed as long
as my contact information is left in place.

Let me know, if you have suggestions or see something that can
be done in a better way, so I might learn from it. Thanks!
*/

// Includes
#include <stdint.h>
#include <stdlib.h>
#include <stdarg.h>

#include "cdbq.h"
#include "print.h"
#include "common.h"
#include "cmdline.h"
#include "emoji.h"
#include "str.h"
#include "x86types.h"
#include "ideatapi.h"
#include "biosint.h"


// Defines
#define _CMDLINE_ARG_COUNT   8

//#define DEBUG_MOCK


// Globals

// Command-line stuff
// Usages: /p port /d dev /s speed [/f] [/q] (CMDLINE_USAGE_GRP_2)
//         /p port /d dev /i                 (CMDLINE_USAGE_GRP_3)
//         /p port /d dev /r [/q]            (CMDLINE_USAGE_GRP_4)
//         /?

cmdline_arg_t args[_CMDLINE_ARG_COUNT] =
{
  {.arg = '?', .flags = CMDLINE_ARG_HELP, .usage_group =
    CMDLINE_USAGE_GROUP_1 },
  {.arg = 'p', .flags = CMDLINE_ARG_HAS_OPT, .usage_group =
    CMDLINE_USAGE_GROUP_2 | CMDLINE_USAGE_GROUP_3 | CMDLINE_USAGE_GROUP_4},
  {.arg = 'd', .flags = CMDLINE_ARG_HAS_OPT, .usage_group =
    CMDLINE_USAGE_GROUP_2 | CMDLINE_USAGE_GROUP_3 | CMDLINE_USAGE_GROUP_4},
  {.arg = 's', .flags = CMDLINE_ARG_HAS_OPT, .usage_group =
    CMDLINE_USAGE_GROUP_2, },
  {.arg = 'i', .flags = CMDLINE_ARG_NONE, .usage_group =
    CMDLINE_USAGE_GROUP_3, },
  {.arg = 'r', .flags = CMDLINE_ARG_NONE, .usage_group =
    CMDLINE_USAGE_GROUP_4, },
  {.arg = 'f', .flags = CMDLINE_ARG_OPTIONAL, .usage_group =
    CMDLINE_USAGE_GROUP_2, },
  {.arg = 'q', .flags = CMDLINE_ARG_QUIET, .usage_group =
    CMDLINE_USAGE_GROUP_2 | CMDLINE_USAGE_GROUP_4 },
};

cmdline_arg_t* arg_help = &(args[0]);
cmdline_arg_t* arg_port = &(args[1]);
cmdline_arg_t* arg_dev = &(args[2]);
cmdline_arg_t* arg_speed = &(args[3]);
cmdline_arg_t* arg_info = &(args[4]);
cmdline_arg_t* arg_reset = &(args[5]);
cmdline_arg_t* arg_force = &(args[6]);
cmdline_arg_t* arg_quiet = &(args[7]);

const char* str_dev =" device...";
const char* str_cmd_ok = " ok.";
const char* str_cmd_err = " not ok.";

// Converted command-line argument values
ioport_t port_prm;
ide_dev_t dev_prm;
int rate_prm;

const uint8_t _CDBQ_SUCCESS_EMOJI_COUNT = 11; // Keep in sync with list below!

uint16_t cdbq_success_emojis[] =
{
  EMJ437_OUTLINED_SMILEY | (BIOSINT_10h_CHATTR_FCOL_YELLOW << 8),
  EMJ437_OPAQUE_SMILEY | (BIOSINT_10h_CHATTR_FCOL_YELLOW << 8),
  EMJ437_HEART |
    ((BIOSINT_10h_CHATTR_FCOL_LIGHTRED | BIOSINT_10h_CHATTR_BLINK) << 8),
  EMJ437_DIAMOND |
    ((BIOSINT_10h_CHATTR_FCOL_LIGHTCYAN | BIOSINT_10h_CHATTR_BLINK) << 8),
  EMJ437_CLUBS | (BIOSINT_10h_CHATTR_FCOL_WHITE << 8),
  EMJ437_SPADES | (BIOSINT_10h_CHATTR_FCOL_WHITE << 8),
  EMJ437_8TH_NOTE | (BIOSINT_10h_CHATTR_FCOL_CYAN << 8),
  EMJ437_2_MUSICAL_NOTES | (BIOSINT_10h_CHATTR_FCOL_MAGENTA << 8),
  EMJ437_SOLAR_SYMBOL  | (BIOSINT_10h_CHATTR_FCOL_YELLOW << 8),
  EMJ437_DOUBLE_EXCL_MARK  | (BIOSINT_10h_CHATTR_FCOL_RED << 8),
  EMJ437_FAKE_MONEY_SIGN  | (BIOSINT_10h_CHATTR_FCOL_LIGHTGREEN << 8)
};


// Returns true, if error was handled, false otherwise.
void cdbq_print_ide_error(ide_error_t err)
{
  char *err_text;

  if (err == E_IDE_SUCCESS)
    return;

  err_text = ide_get_error_text(err);
  eprintf("ERROR: %s\n", err_text);
}

// Returns true on success, false if error text returned
bool cdbq_get_error_text(int err_code, char* err_text_buf, ...)
{
  switch (err_code)
  {
    case E_CDBQ_SUCCESS:
      return true;

    case E_CDBQ_OPT_INVALID:
    {
      va_list args;
      va_start(args, err_text_buf);
      vsprintf(err_text_buf, "Option for argument '%c' is invalid.", args);
      va_end(args);
      return false;
    }

    case E_CDBQ_OPT_WRONG_SPEED:
    {
      va_list args;
      va_start(args, err_text_buf);
      vsprintf(err_text_buf,
        "Speed parameter can only range from 0 to %u kb/s.", args);
      va_end(args);
      return false;
    }
  }

  sprintf(err_text_buf, __FILE__ ": Unknown error. Code: %d", err_code);
  return false;
}

// Returns true on success, otherwise false.
bool cdbq_cmdline_parse(cmdline_arg_t* arg, char* appspec_err_text_buf)
{
  // Convert switches with options

  switch (arg->arg)
  {
    case 'p':
      port_prm = (ioport_t)atoi(arg->opt_val);

      if (!port_prm)
        return cdbq_get_error_text(E_CDBQ_OPT_INVALID, appspec_err_text_buf,
          arg->arg);

      switch(port_prm)
      {
        case 1:
          port_prm = IDE1_BASE;
          break;
        case 2:
          port_prm = IDE2_BASE;
          break;
        case 3:
          port_prm = IDE3_BASE;
          break;
        case 4:
          port_prm = IDE4_BASE;
          break;
      }

      break;

    case 'd':
    {
      char dev_chr = to_lower(arg->opt_val[0]);

      switch(dev_chr)
      {
        case 'm':
          dev_prm = IDE_DEV_MASTER;
          break;
        case 's':
          dev_prm = IDE_DEV_SLAVE;
          break;
        default:
          return cdbq_get_error_text(E_CDBQ_OPT_INVALID, appspec_err_text_buf,
            arg->arg);
      }
    } break;

    case 's':
      rate_prm = atoi(arg->opt_val);

      // Check if rate param is correct
      if (rate_prm > CDBQ_CFG_MAX_SPEED_SETTING)
        return cdbq_get_error_text(E_CDBQ_OPT_WRONG_SPEED, appspec_err_text_buf,
          CDBQ_CFG_MAX_SPEED_SETTING);
      break;
  }

  return true;
}


void cdbq_emoji_init_srand(char* s1, char* s2)
{
  unsigned int seed = 0;
  int i;

  i = 0;
  while(s1[i] != '\0') seed += (unsigned int)(s1[i++]);

  i = 0;
  while(s2[i] != '\0') seed += (unsigned int)(s2[i++]);

  srand(seed);
}

void cdbq_print_success_emoji(char *s1, char *s2) // :-)
{
  int i;
  bool is_color;
  uint16_t c;

  unsigned char mode = biosint_video_get_mode();

  if (mode != BIOSINT_10h_00h_AL_80x25_TEXT_COLOR &&
    mode != BIOSINT_10h_00h_AL_80x25_TEXT_MONO)
    return; // Emojis not supported in strange modes.

  is_color = mode == BIOSINT_10h_00h_AL_80x25_TEXT_COLOR;

  cdbq_emoji_init_srand(s1, s2);
  i = rand() % (_CDBQ_SUCCESS_EMOJI_COUNT);

  c = cdbq_success_emojis[i];
  fflush(stdout);

  // Strip attribute stuff in mono mode
  biosint_sc_video_write_char_attr(is_color ? c : c & 0xff);
}

inline void cdbq_display_drive_info(ide_drive_info_t *di)
{
  printf("  Model name    : %s\n"
         "  Firmware rev. : %s\n",
    ltrim(di->model_num), ltrim(di->fw_rev));
}

int main(const int argc, const char *argv[])
{
  ide_error_t ide_error;
  char ide_str[20]; // e.g. "quarternary master" (longest)
  ide_drive_info_t di;
  unsigned int packet[6]; // ATAPI packet command
  bool loud;
  int ret;

  ret = cmdline_init(_CMDLINE_ARG_COUNT, args, argc, argv,
    cdbq_cmdline_parse);

  if (ret == E_CMDLINE_INIT_USAGESHOWN)
    return COMMON_EC_SUCCESS;
  else if (ret != E_CMDLINE_INIT_SUCCESS)
    return COMMON_EC_BAD_CMDLINE;

  // Quiet mode?
  loud = !(arg_quiet->flags & CMDLINE_ARG_FOUND);

   // Turn off buffering, updates text faster.
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  if (loud)
  {
    ide_str[0] = '\0';
    switch(port_prm)
    {
      case IDE1_BASE:
        strcat(ide_str, "primary");
        break;
      case IDE2_BASE:
        strcat(ide_str, "secondary");
        break;
      case IDE3_BASE:
        strcat(ide_str, "tertiary");
        break;

      case IDE4_BASE:
        strcat(ide_str, "quarternary");
        break;
      default:
        sprintf(ide_str, "port_prm 0x%x", port_prm);
    }

    if (dev_prm == IDE_DEV_MASTER)
      strcat(ide_str, " master");
    else
      strcat(ide_str, " slave");

    // Read drive information
    PRINT("\nReading drive information from ");
    PRINT(ide_str);
    PRINT(str_dev);
  }

  ide_error = ide_get_atapi_info(port_prm, dev_prm, &di);

#ifdef DEBUG_MOCK
  ide_error = E_IDE_SUCCESS;
  sprintf(di.model_num, "Mock device");
  sprintf(di.fw_rev, "Rev. 42");
#endif

  if (ide_error != E_IDE_SUCCESS)
  {
    if (loud) PRINTLN(str_cmd_err);
    cdbq_print_ide_error(ide_error);

    return CDBQ_EC_IDE_ERROR;
  }

  // Check device, if not forced.
  if (!(arg_force->flags & CMDLINE_ARG_FOUND))
  {
    ide_error_t ide_error = ide_check_if_cdrom(&di);

#ifdef DEBUG_MOCK
    ide_error = E_IDE_SUCCESS;
#endif

    if (ide_error != E_IDE_SUCCESS)
    {
      if (loud)
      {
        PRINTLN(str_cmd_err);
        cdbq_display_drive_info(&di); // Still the info we got, so print it!
      }

      cdbq_print_ide_error(ide_error);

      return CDBQ_EC_IDE_ERROR;
    }
  }

  if (loud)
  {
    PRINTLN(str_cmd_ok);
    cdbq_display_drive_info(&di);
  }

  if (arg_info->flags & CMDLINE_ARG_FOUND)
    return COMMON_EC_SUCCESS; // We are done here!

  // Process reset command
  if (arg_reset->flags & CMDLINE_ARG_FOUND)
  {
    // Issue soft-reset command and quit.
    if (loud)
    {
      PRINT("Sending soft-reset to ");
      PRINT(ide_str);
      PRINT(str_dev);
    }

    ide_error = ide_atapi_soft_reset(port_prm, dev_prm);

#ifdef DEBUG_MOCK
    ide_error = E_IDE_SUCCESS;
#endif

    if (ide_error != E_IDE_SUCCESS)
    {
      if (loud) PRINTLN(str_cmd_err);
      cdbq_print_ide_error(ide_error);
      return CDBQ_EC_IDE_ERROR;
    }

    if (loud) PRINTLN(str_cmd_ok);

    return COMMON_EC_SUCCESS; // We are done here!
  }

  if (loud)
  {
    if (rate_prm == 0 || rate_prm > CDBQ_CFG_MAX_SPEED_SETTING)
      PRINT("Setting maximum CD-ROM read rate on ");
    else
      printf("Setting CD-ROM read rate of %u kb/s on ", rate_prm);

    PRINT(ide_str);
    PRINT(str_dev);
  }

  // Prepare and send ATAPI packet for command "SET CD SPEED".
  ide_atapi_get_cd_speed_packet(rate_prm, packet);
  ide_error = ide_atapi_send_packet_cmd(port_prm, dev_prm, packet);

#ifdef DEBUG_MOCK
    ide_error = E_IDE_SUCCESS;
#endif

  if (ide_error != E_IDE_SUCCESS)
  {
    unsigned int ide_error_bits;

    if (loud) PRINTLN(str_cmd_err);
    cdbq_print_ide_error(ide_error);

    ide_error_bits = ide_get_error_reg_bits();
    eprintf("IDE error register value was: 0x%xh\n", ide_error_bits);

    return CDBQ_EC_IDE_ERROR;
  }

  if (loud) PRINT(str_cmd_ok);

  if (loud && rate_prm != 0 && rate_prm < CDBQ_CFG_MAX_SPEED_SETTING)
  {
    PRINT("\nEnjoy the silence! ");
    cdbq_print_success_emoji(di.model_num, di.fw_rev);
  }

  return COMMON_EC_SUCCESS;
}
