/*
Copyright (C) 2024 Christoph Boelitz

E-mail: tools@dosforum.de
Project page: http://tools.dosforum.de/

This program is free software as long as its used privately.
Commercial use, as unlikely as it may seem, is hereby prohibited.
The source code may be modified and distributed as long
as my contact information is left in place.

Let me know, if you have suggestions or see something that can
be done in a better way, so I might learn from it. Thanks!
*/

// Includes
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "COMMON.H"
#include "CMDLINE.H"
#include "STR.H"
#include "PRINT.H"

// Macros
#define _HAS_SW_CH(c) ((c == '/') || (c == '-')) // Has switch char?

struct { int code; const char* text; } _cmdline_error_texts[] =
  { FOREACH_CMDLINE_ERROR(GENERATE_STRUCT_MEMBER) };


// Globals
cmdline_arg_t* _help_arg = NULL;
cmdline_arg_t* _quiet_arg = NULL;

// Implementation

const char* cmdline_get_error_text(cmdline_error_t error)
{
  int i;

  for (i=0; i <= (int)E_CMDLINE_UNKNOWN; i++)
    if (_cmdline_error_texts[i].code == (int)error)
      return _cmdline_error_texts[i].text;

  // Should not happen, but ok...
  return __FILE__ ": No error description found.";
}

void _cmdline_print_args(int arg_count, cmdline_arg_t args[])
{
  int i;

  for (i = 0; i < arg_count; i++)
    printf("arg %c, found %d, has_opt %d, opt_val |%s|\n",
      args[i].arg,
      (args[i].flags & CMDLINE_ARG_FOUND) > 0,
      (args[i].flags & CMDLINE_ARG_HAS_OPT) > 0,
      args[i].opt_val);
}


int cmdline_parse(int arg_count, cmdline_arg_t args[], int argc,
  const char* argv[], char* appspec_err_txt_buf,
  cmdline_parse_appspec_fptr_t parse_appspec_fptr)
{
  int argv_idx, arg_idx;
  int valid_grp_mask, cur_grp_mask;
  int seen_grps = CMDLINE_USAGE_GROUP_NONE;

  _help_arg = NULL;
  _quiet_arg = NULL;

  for (argv_idx = 1; argv_idx < argc; argv_idx++)
  {
    const char* cur_argv;
    int cur_argv_len;
    char cur_argv_l;
    bool is_sw;

    cur_argv = argv[argv_idx];

check:
    cur_argv_len = strlen(cur_argv);

    if (cur_argv_len < 1)
      return E_CMDLINE_UNKNOWN; // Empty argument, should never occur


    is_sw = cur_argv_len > 1 && _HAS_SW_CH(cur_argv[0]);

    if (!is_sw)
      return E_CMDLINE_UNEXPECTED_OPTION;

    cur_argv_l = to_lower(cur_argv[1]);

    for (arg_idx = 0; arg_idx < arg_count; arg_idx++)
    {
      cmdline_arg_t* cur_arg = &(args[arg_idx]);

      // Update seen groups mask
      seen_grps |= cur_arg->usage_group;

      // Ok, we found a switch-arg, but is it the one?

      if (cur_argv_l != cur_arg->arg)
        continue;

      // Found it, but was it found before?
      if (cur_arg->flags & CMDLINE_ARG_FOUND)
        return E_CMDLINE_DUPLICATE_ARGS;

      cur_arg->flags |= CMDLINE_ARG_FOUND;

      if (!(cur_arg->flags & CMDLINE_ARG_HAS_OPT))
      {
        // Check, if another switch-arg is concatenated to this argv
        if (cur_argv_len >= 3)
        {
          cur_argv += 2;
          goto check;
        }

        break;
      }

      // So it has an option!
      // Is it part of the current argv or next?

      if (strlen(cur_argv) > 2 && cur_argv[2] != ' ')
      {
        cur_arg->opt_val = cur_argv + 2;
      }
      else
      {
        // It's in next argv, so look-ahead!
        const char* next_argv;

        // But maybe it's missing?
        if (argv_idx+1 >= argc)
          return E_CMDLINE_MISSING_OPTION;

        next_argv = argv[argv_idx + 1];

        // If next argv is a switch, the option is missing!
        if (strlen(next_argv) == 2 && _HAS_SW_CH(next_argv[0]))
          return E_CMDLINE_MISSING_OPTION;

        // Cannot be switch (would have exited above),
        // which means we found the option-arg, so save the value.

        cur_arg->opt_val = next_argv;

        // Since we handled the option in next argv,
        // skip over it in next outer loop.
        argv_idx++;
      }

      // Call app-specific parsing function
      if (parse_appspec_fptr != NULL &&
        !parse_appspec_fptr(cur_arg, appspec_err_txt_buf))
      {
        // Check, if app-spec error function did not provide a text
        if (is_null_or_empty(appspec_err_txt_buf))
          return E_CMDLINE_APP_UNSPECIFIC;

        return E_CMDLINE_APP_SPECIFIC;
      }

      break;
    }

    // If inner loop completed, it's an unknown switch
    if (arg_idx >= arg_count)
      return E_CMDLINE_UNEXPECTED_ARG;
  }

  valid_grp_mask = 0;
  for (cur_grp_mask = CMDLINE_USAGE_GROUP_1;
    cur_grp_mask <= CMDLINE_USAGE_GROUP_8;
    cur_grp_mask <<= 1)
  {
    int reqd_found_cnt, reqd_cnt;

    // Skip if not seen before
    if (!(seen_grps & cur_grp_mask))
      continue;

    reqd_found_cnt = 0;
    reqd_cnt = 0;
    for (arg_idx = 0; arg_idx < arg_count; arg_idx++)
    {
      bool found, has_valid_grp, has_other_grp;
      cmdline_arg_t* cur_arg = &(args[arg_idx]);

      // Look for "help" arg
      if (_help_arg == NULL && cur_arg->flags & CMDLINE_ARG_HELP)
        _help_arg = cur_arg;

      // Look for "quiet" arg
      if (_quiet_arg == NULL && cur_arg->flags & CMDLINE_ARG_QUIET)
      {
        _quiet_arg = cur_arg;
        _quiet_arg->flags |= CMDLINE_ARG_OPTIONAL;
      }

      found = cur_arg->flags & CMDLINE_ARG_FOUND;
      has_valid_grp = cur_arg->usage_group & valid_grp_mask;
      has_other_grp = !has_valid_grp && cur_arg->usage_group > 0;

      if (found && valid_grp_mask && has_other_grp)
        return E_CMDLINE_ARG_CONFLICT;

      if (cur_arg->usage_group & cur_grp_mask && 
        (!(cur_arg->flags & CMDLINE_ARG_OPTIONAL)) )
      {
        reqd_cnt++;

        if (found)
          reqd_found_cnt++;
      }
    }

    if (reqd_cnt > 0 && reqd_found_cnt >= reqd_cnt)
      valid_grp_mask = cur_grp_mask;
  }

  if (!valid_grp_mask)
    return E_CMDLINE_MISSING_ARG;


  return E_CMDLINE_SUCCESS;
}

// Call this at the start of your program, if you want to use handle parsing
// of command-line arguments and also printing of errors altogether.
int cmdline_init(int arg_count, cmdline_arg_t args[], const int argc,
  const char* argv[], cmdline_parse_appspec_fptr_t parse_appspec_fn_ptr)
{
  // Do command-line parsing...

  cmdline_error_t cmdline_error;
  char appspec_err_txt_buf[200];

  // Easiest case first: No arguments given, show usage
  if (argc < 2)
  {
#ifdef COMMON_SHOW_BANNER_ON_STARTUP
    common_show_banner();
    PRINTNLNL();
#endif

#ifdef COMMON_USAGE_TEXT
  common_print_usage();
#else
  return E_CMDLINE_INIT_BAD_ARGS;
#endif

    return E_CMDLINE_INIT_USAGESHOWN;
  }

  // Don't show banner yet! Quiet switch might have been used!
  cmdline_error = cmdline_parse(arg_count, args, argc, argv,
    appspec_err_txt_buf, parse_appspec_fn_ptr);

  // Show errors, if there were any.
  if (cmdline_error != E_CMDLINE_SUCCESS)
  {
    const char* err_txt = appspec_err_txt_buf;

#ifdef COMMON_SHOW_BANNER_ON_STARTUP
    // On error show banner regardless of quiet-switch
    common_show_banner();
    PRINTNLNL();
#endif

    if (cmdline_error != E_CMDLINE_APP_SPECIFIC)
      err_txt = cmdline_get_error_text(cmdline_error);

    eprintf("ERROR: %s", err_txt);

    return E_CMDLINE_INIT_BAD_ARGS;
  }

  // From here on cmdline_parse() did succeed, and it needs to have completed
  // in order to use _help_arg!

  // Check for help argument
#if defined(COMMON_USAGE_TEXT)
  if (_help_arg != NULL && (_help_arg->flags & CMDLINE_ARG_FOUND))
  {
    // Usage is always shown on error, if requested,
    // no matter if quiet-switch detected!
#ifdef COMMON_SHOW_BANNER_ON_STARTUP
    common_show_banner();
    PRINTNLNL();
#endif

    common_print_usage();

    return E_CMDLINE_INIT_USAGESHOWN;
  }
#endif

// Arguments correct, no help-switch detected, finally show banner if requested
#ifdef COMMON_SHOW_BANNER_ON_STARTUP
  if (_quiet_arg == NULL || !(_quiet_arg->flags & CMDLINE_ARG_FOUND))
    common_show_banner();
#endif

  return E_CMDLINE_INIT_SUCCESS;
}


/* This stuff is deprecated. Will be gone in a future version.

inline bool _is_arg_found(int arg_count, cmdline_arg_t args[], char arg_chr)
{
  int i;

  for (i = 0; i < arg_count; i++)
    if (args[i].arg == arg_chr)
      return args->flags & CMDLINE_ARG_FOUND;

  return false;
}

// Returns true, if success, false on error
bool _cmdline_parse_groups(int arg_count, cmdline_arg_t args[],
  const char* grps_str, bool conflict_mode)
{
  int chr_idx;
  int found_cnt;
  int grp_start_ofs;
  int cur_grp_len;

  if (is_null_or_empty(grps_str))
    return true;

  found_cnt = 0;
  grp_start_ofs = 0;
  chr_idx = 0;
  cur_grp_len = -1;
  do
  {
    char cur_grp_chr;

    cur_grp_chr = grps_str[grp_start_ofs+chr_idx];

    if (cur_grp_chr == '|' || cur_grp_chr == '\0')
    {
      // End of group reached

      cur_grp_len = chr_idx;

      // Check if all found in completed group
      if (cur_grp_len > 0 && cur_grp_len == found_cnt)
        return true;

      // Check, if we are at end of string. If so, return according to mode:
      // In conflict mode, we return true, because no conflicts are found.
      // In mandatory-check mode, we return false, because option was missing.
      if (cur_grp_chr == '\0')
        return conflict_mode;

      // Setup for next group
      grp_start_ofs += chr_idx+1;
      chr_idx = -1;
      found_cnt = 0;
    }
    else if (_is_arg_found(arg_count, args, cur_grp_chr))
    {
      // In conflict-mode, more than one found option
      // per group is not allowed.
      if (conflict_mode && found_cnt > 0)
        return false;

      found_cnt++;
    }

    chr_idx++;
  } while (true);

  // This code should never be reached.
  assert(false);
}
*/
