#include <stdio.h>
/*
Copyright (C) 2024 Christoph Boelitz

E-mail: tools@dosforum.de
Project page: http://tools.dosforum.de/

This program is free software as long as its used privately.
Commercial use, as unlikely as it may seem, is hereby prohibited.
The source code may be modified and distributed as long
as my contact information is left in place.

Let me know, if you have suggestions or see something that can
be done in a better way, so I might learn from it. Thanks!
*/

#include "ideatapi.h"

// Generate error description table

// Holds error register values in case of error
unsigned _ide_error_reg_bits = 0;

struct { int code; char* text; } _ide_error_texts[] =
  { FOREACH_IDE_ERROR(GENERATE_STRUCT_MEMBER) };

char* ide_get_error_text(ide_error_t error)
{
  int i;

  for (i=0; i <= (int)E_IDE_UNKNOWN; i++)
    if (_ide_error_texts[i].code == (int)error)
      return _ide_error_texts[i].text;

  // Should not happen, but ok...
  return __FILE__ ": No error description found.";
}

inline unsigned int _ide_swap_bytes(unsigned int w)
{
  char b1 = w >> 8;
  char b2 = w - (b1 << 8);
  return (b2 << 8)+ b1;
}

// Wait one I/O cycle by reading the alternate status register.
inline void _ide_waste_io_cycle(ioport_t port)
{
  inp(port+IDE_ALT_STATUS);
}


// Returns error code. Be sure to re-enable() interrupts in calling code
// when this function executes successful.
ide_error_t _ide_send_ata_cmd(ioport_t port, ide_dev_t dev,
  uint8_t opcode, bool wait_busy, bool wait_ready, bool wait_drq)
{
  /*
    Sends ATAPI command:
      1. Poll status until BSY=0 (optional via wait_busy)
      2. Disable interrupts (clear bit NIEN)
      3. Poll status until DRDY=1 (optional via wait_ready)
      4. select device (master/slave)
      Only for PACKET command:
        5. Set IEN=0 in device control register (0=enable ints)
      6. send ATAPI command (opcode)
      7. Wait a bit
      8. Poll status until DRQ=1 (optional via wait_drq)

    Note: This command disables interrupts.
  */

  int ret; //Used for various return values

  // 1
  if (wait_busy)
  {
    ret = ide_poll_register(port+IDE_STATUS, IDE_STATUS_BUSY, 0);

    if (ret != E_IDE_SUCCESS)
      return E_IDE_WAITBUSY;
  }

  // 2
  _disable();

  // 3
  if (wait_ready)
  {
    ret = ide_poll_register(port+IDE_STATUS, IDE_STATUS_DRDY, IDE_STATUS_DRDY);

    if (ret != E_IDE_SUCCESS)
    {
      _enable();
      return E_IDE_WAITDRDY;
    }
  }

  // 4
  outp(port+IDE_DRIVE_HEAD, dev);

  // 5
  if (opcode == IDE_ATAPI_PACKET_COMMAND)
    outp(port+IDE_DEVCTRL, IDE_DEVCTRL_NIEN); //0000 0010

  // 6
  outp(port+IDE_STATUS, opcode);

  // 7 For ATAPI packet command we should wait 400ns,
  //   but we'll wait a bit longer.
  delay(2); // Probably overkill, but will do the trick.

  // 8
  if (wait_drq)
  {
    ret = ide_poll_register(port+IDE_STATUS, IDE_STATUS_DRQ, IDE_STATUS_DRQ);

    if (ret != E_IDE_SUCCESS)
    {
      _enable();
      return E_IDE_WAITDRQ;
    }
  }

  return E_IDE_SUCCESS;
}

unsigned ide_get_error_reg_bits()
{
  return _ide_error_reg_bits;
}

ide_error_t ide_check_if_cdrom(ide_drive_info_t* di)
{
  // Check, if device speaks ATAPI

  if
  (
    (di->general_config & IDE_ATAPI_GEN_CFG_WORD_PROTOTYPE)
    != IDE_ATAPI_GEN_CFG_WORD_PROTOTYPE_ATAPI
  ) // 11000000
    return E_IDE_NOT_ATAPI_DEV;

  // Check, if device is a CD-ROM device
  if
  (
    (di->general_config & IDE_ATAPI_GEN_CFG_WORD_DEVTYPE)
    != IDE_ATAPI_GEN_CFG_WORD_DEVTYPE_CDROM
  )
    return E_IDE_NOT_A_CDROM;

  // Check, if device accepts 12-byte ATAPI packets
  // (All CD-ROMs according to spec)

  if
  (
    (di->general_config & IDE_ATAPI_GEN_CFG_WORD_PKTSIZE)
    != IDE_ATAPI_GEN_CFG_WORD_PKTSIZE_12B
  )
    return E_IDE_WRONG_PKT_SIZE;

  return E_IDE_SUCCESS;
}


ide_error_t ide_get_atapi_info(ioport_t port, ide_dev_t dev,
  ide_drive_info_t *di)
{
  /*
    1. Send IDENTIFY DRIVE ATA command; Drive will
       abort and reveal ATAPI signature.
    2. Read ATAPI signature.
    3. Send IDENTIFY DEVICE ATAPI command; Drive will
       store information in sector buffer.
    4. Read information from sector buffer and store
       in drive_info struct.

    Note: According to spec the IDENTIFY DRIVE(!) ATA command shall be
    aborted by CD-ROM drives, so as not to confuse them with ATA drives.
  */

  ide_error_t ide_error;
  unsigned id_drive_data[256];
  unsigned ide_cyl_high, ide_cyl_low;
  int i;

  // 1
  ide_error = _ide_send_ata_cmd(port, dev, IDE_IDENTIFY_DRIVE,
    true, true, false);

  if (ide_error != E_IDE_SUCCESS)
    return ide_error;

  ide_error = ide_check_error_condition(port);

  _enable();

  // At this point, if the device is an ATAPI device, it should have aborted
  // the previous command.

  if (ide_error != E_IDE_COMMAND_ABORTED)
    return E_IDE_IDDRIVE_NOT_ABORTED;

  // 2
  ide_cyl_high = inp(port+IDE_CYL_HIGH);
  ide_cyl_low = inp(port+IDE_CYL_LOW);

  if
  (
    (ide_cyl_high == ATAPI_SIG_CYL_HIGH) &&
    (ide_cyl_low  == ATAPI_SIG_CYL_LOW)
  )
    return E_IDE_NO_ATAPI_SIG;

  // 3 (Spec says: "Shall be accepted regardless of DRDY state")
  ide_error = _ide_send_ata_cmd(port, dev, IDE_ATAPI_IDENTIFY_DEVICE,
    true, false, true);

  if (ide_error != E_IDE_SUCCESS)
    return ide_error;

  // 4
  for (i=0; i<256; i++)
  id_drive_data[i] = inpw(port+IDE_DATA);

  _enable();

  // Copy model number (offset 27-46)
  for (i=27;i<=46;i++)
  {
    di->model_num[(i-27)*2] = (char)(id_drive_data[i] >> 8);
    di->model_num[(i-27)*2+1] = (char)id_drive_data[i];
  }

  di->model_num[sizeof(di->model_num)-1] = '\0';

  // Copy firmware revision (offset 23-26)
  for (i=23;i<=26;i++)
  {
    di->fw_rev[(i-23)*2] = (char)(id_drive_data[i] >> 8);
    di->fw_rev[(i-23)*2+1] = (char)id_drive_data[i];
  }

  di->fw_rev[sizeof(di->fw_rev)-1] = '\0';
  di->general_config = id_drive_data[0]; //One word, offset 0

  return E_IDE_SUCCESS;
}

ide_error_t ide_poll_register(ioport_t port, unsigned mask, unsigned state)
{
  // Polls IDE register until bits set in mask match their value in state
  // Specify bits that should be tested in mask by setting them high and
  // their respective value in state.

  // Note: Be sure to specify the register port, not the IDE base port.

  unsigned reg_bits;
  int i;

  reg_bits = inp(port);

  i=0;
  while ((reg_bits & mask) != state)
  {
    delay(100);

    i+=100;

    if (i >= 5000) // try ~5 seconds...
      return E_IDE_TIMEOUT;

    reg_bits = inp(port);
  }

  return E_IDE_SUCCESS;
}

ide_error_t ide_check_error_condition(ioport_t port)
{
  // Checks IDE error regsiter for error and returns type of error.

  // If this function is called right after _ide_send_ata_cmd() it seems we
  // have to give the controller some time to validate the error register,
  // so we wait for BSY=0 here.

  ide_error_t err;
  unsigned status_bits;

  _ide_error_reg_bits = 0;

  err = ide_poll_register(port+IDE_STATUS, IDE_STATUS_BUSY, 0);

  if (err != E_IDE_SUCCESS)
    return E_IDE_WAITBUSY;

  status_bits = inp(port+IDE_STATUS);

  if ((status_bits & IDE_STATUS_ERR) == 0)
    return E_IDE_SUCCESS;

  //Else: There has been an error; Retrieve error bits from error register
  //      to investigate further. Also, since the ERR bit has been set in
  //      status register, we can assume that the error register is valid.

  _ide_error_reg_bits = inp(port+IDE_ERROR);

  if (((_ide_error_reg_bits) & IDE_ERROR_ABRT) == IDE_ERROR_ABRT)
    return E_IDE_COMMAND_ABORTED;

  // Spec defines more errors, but they are not of interest here.
  return E_IDE_UNKNOWN;
}

void ide_atapi_get_cd_speed_packet(unsigned speed, unsigned packet[])
{
  // Fills an array of words that represent the
  // SET CD SPEED ATAPI command packet with values.

  // Word 1 has qualifier in reserved portion, add  support for Pioneer?
  packet[0] = IDE_ATAPI_PKT_SET_CDROM_SPEED;
  packet[1] = speed==0 ? 0xffff : _ide_swap_bytes(speed); // 0xffff = max speed
  packet[2] = 0; // This actually holds write-speed, maybe put 0xffff here?
  packet[3] = 0;
  packet[4] = 0;
  packet[5] = 0;
}

ide_error_t ide_atapi_soft_reset(ioport_t port, ide_dev_t dev)
{
  // Sends soft-reset command to IDE device (experimental)

  int ret;

  // Spec says, shall be accepted when BSY or DRQ is high, DRDY is low
  ret = _ide_send_ata_cmd(port, dev, IDE_ATAPI_SOFTRESET, 0, 0, 0);

  if (ret != E_IDE_SUCCESS)
    return ret;

  // If supported this command will not end in an error condition, otherwise
  // it will probably result in command aborted state.
  ret = ide_check_error_condition(port);

  _enable();

  if (ret == E_IDE_COMMAND_ABORTED)
    return E_IDE_ATAPI_SRST_ABORTED;

  return ret;
}

ide_error_t ide_atapi_send_packet_cmd(ioport_t port, ide_dev_t dev,
  unsigned packet[])
{
  ide_error_t err; // Used for various return values
  int i;

  err = _ide_send_ata_cmd(port, dev, IDE_ATAPI_PACKET_COMMAND,
    true, true, true);

  if (err != E_IDE_SUCCESS)
    return err;

  delay(5);

  // Wait for BSY=0
  err = ide_poll_register(port+IDE_STATUS, IDE_STATUS_BUSY, 0);

  if (err != E_IDE_SUCCESS)
    return E_IDE_WAITBUSY;

  // Sending the packet command is done by 6 word writes. These
  // make up the 12 bytes of the command's size.

  for(i=0; i<6; i++)
  {
    // outportw(), because we send a word (16 bit)
    outpw(port+IDE_DATA, packet[i]);

    // Wait one I/O cycle by reading the alternate status register.
    _ide_waste_io_cycle(port);
  }

  // Check alternative status register one more time to waste one i/o cycle.
  _ide_waste_io_cycle(port);

  err = ide_check_error_condition(port);

  _enable();

  return err;
}
