/*
Copyright (C) 2024 Christoph Boelitz

E-mail: tools@dosforum.de
Project page: http://tools.dosforum.de/

This program is free software as long as its used privately.
Commercial use, as unlikely as it may seem, is hereby prohibited.
The source code may be modified and distributed as long
as my contact information is left in place.

Let me know, if you have suggestions or see something that can
be done in a better way, so I might learn from it. Thanks!
*/

// Module for calling BIOS interrupts
// (This is still very preliminary...)

#include "biosint.h"


// Some handy shortcut functions

// Advance cursor pos by count (won't wrap!)
void biosint_sc_video_adv_cur(int count)
{
  unsigned char l;
  unsigned char c;

  biosint_video_get_cursor_pos(&l, &c);
  biosint_video_set_cursor_pos(l, c+count);
}

// Writes char with encoded attribute to screen and advances cursor pos.
void biosint_sc_video_write_char_attr(unsigned int c_attr)
{
  biosint_video_09h_write_char((unsigned char)(c_attr), c_attr >> 8);
  biosint_sc_video_adv_cur(1);
}

// And int86() wrapping

unsigned char biosint_video_get_mode()
{
  union REGS in_regs;
  union REGS out_regs;

  in_regs.h.ah = BIOSINT_10h_0fh_GET_MODE;

  int86(BIOSINT_10h_VIDEO, &in_regs, &out_regs);

  return out_regs.h.al;
}

void biosint_video_set_mode(unsigned char mode)
{
  union REGS in_regs;
  union REGS out_regs;

  in_regs.h.ah = BIOSINT_10h_00h_SET_MODE;
  in_regs.h.al = mode;    // ASCII code of char

  int86(BIOSINT_10h_VIDEO, &in_regs, &out_regs);
}

void biosint_video_set_cursor_pos(unsigned char line, unsigned char col)
{
  union REGS in_regs;
  union REGS out_regs;

  in_regs.h.ah = BIOSINT_10h_02h_SET_CURSOR_POS;
  in_regs.h.bh = 0;   // Display page no.
  in_regs.h.dh = line;
  in_regs.h.dl = col;

  int86(BIOSINT_10h_VIDEO, &in_regs, &out_regs);
}

void biosint_video_get_cursor_pos(unsigned char *line, unsigned char *col)
{
  union REGS in_regs;
  union REGS out_regs;

  in_regs.h.ah = BIOSINT_10h_03h_GET_CURSOR_POS;
  in_regs.h.bh = 0;   // Display page no.

  int86(BIOSINT_10h_VIDEO, &in_regs, &out_regs);

  *line = out_regs.h.dh;
  *col = out_regs.h.dl;
}

// Write char/attribute. Does not advance cursor pos.
// Does not filter out control codes such as bell.
void biosint_video_09h_write_char(unsigned char c, unsigned char attr)
{
  union REGS in_regs;
  union REGS out_regs;

  in_regs.h.ah = BIOSINT_10h_09h_WRITE_CHAR;
  in_regs.h.al = c;          // ASCII code of char
  in_regs.h.bl = attr;        // Color?
  in_regs.h.bh = 0;          // Display page no.
  in_regs.w.cx = 1;          // No. of times to write char

  int86(BIOSINT_10h_VIDEO, &in_regs, &out_regs);
}

// Control codes (bell, ...) won't get filtered out.
// Does not advance cursor pos.
void biosint_video_0ah_write_char(unsigned char c)
{
  union REGS in_regs;
  union REGS out_regs;

  in_regs.h.ah = BIOSINT_10h_0ah_WRITE_CHAR;
  in_regs.h.al = c;          // ASCII code of char
  //in_regs.h.bl = 1;        // Color?
  in_regs.h.bh = 0;          // Display page no.
  in_regs.w.cx = 1;          // No. of times to write char

  int86(BIOSINT_10h_VIDEO, &in_regs, &out_regs);
}

// Also advances cursor pos. and executes control codes such as bell symbol.
void biosint_video_0eh_write_char(unsigned char c)
{
  union REGS in_regs;
  union REGS out_regs;

  in_regs.h.ah = BIOSINT_10h_0eh_WRITE_CHAR;
  in_regs.h.al = c;          // ASCII code of char
  in_regs.h.bl = 0;          // Foreground color
  in_regs.w.cx = 1;          // No. of times to write char

  int86(BIOSINT_10h_VIDEO, &in_regs, &out_regs);
}

