/*
Copyright (C) 2024 Christoph Boelitz

E-mail: tools@dosforum.de
Project page: http://tools.dosforum.de/

This program is free software as long as its used privately.
Commercial use, as unlikely as it may seem, is hereby prohibited.
The source code may be modified and distributed as long
as my contact information is left in place.

Let me know, if you have suggestions or see something that can
be done in a better way, so I might learn from it. Thanks!
*/

#ifndef _IDE_H_INCLUDED
#define _IDE_H_INCLUDED

// Includes

#include <stdbool.h>
#include <i86.h>
#include <conio.h>
#include <string.h>

#include "x86types.h"
#include "errmacro.h"

// Defines

// IDE/ATA commands
#define IDE_IDENTIFY_DRIVE             0xec // Also known as IDENTIFY DEVICE

// ATAPI commands
#define IDE_ATAPI_IDENTIFY_DEVICE      0xa1 // a.k.a IDENTIFY PACKET DEVICE
#define IDE_ATAPI_PACKET_COMMAND       0xa0
#define IDE_ATAPI_SOFTRESET            0x08

// ATA register bits
#define IDE_STATUS_ERR                 0x01 // 0000 0001
#define IDE_STATUS_BUSY                0x80 // 1000 0000
#define IDE_STATUS_DRQ                 0x08 // 0000 1000
#define IDE_STATUS_DRDY                0x40 // 0100 0000
#define IDE_ERROR_ABRT                 0x04 // 0000 0100
#define IDE_DEVCTRL_NIEN               0x02 // 0000 0010

// General configuration word (retrieved by IDENTIFY PACKET DEVICE command)
#define IDE_ATAPI_GEN_CFG_WORD_PROTOTYPE 0xc000 // 1100 0000 0000 0000, bitmask
#define IDE_ATAPI_GEN_CFG_WORD_DEVTYPE   0x1f00 // 0001 1111 0000 0000, bitmask
#define IDE_ATAPI_GEN_CFG_WORD_PKTSIZE   0x0003 // 0000 0000 0000 0011, bitmask

// 1000 0000 0000 0000, 10b = ATAPI device
#define IDE_ATAPI_GEN_CFG_WORD_PROTOTYPE_ATAPI 0x8000
// 0000 0101 0000 0000, 05h = CD-ROM device
#define IDE_ATAPI_GEN_CFG_WORD_DEVTYPE_CDROM   0x0500
// 0000 0000 0000 0000, 00b = 12 bytes
#define IDE_ATAPI_GEN_CFG_WORD_PKTSIZE_12B     0x0000

// ATAPI PACKET commands
#define IDE_ATAPI_PKT_SET_CDROM_SPEED  0xbb
#define IDE_ATAPI_PKT_REQUEST_SENSE    0x03

// ATAPI sense-key states
#define IDE_ATAPI_SK_NO_SENSE          0x0
#define IDE_ATAPI_SK_RECOVERED_ERROR   0x1
#define IDE_ATAPI_SK_NOT_READY         0x2
#define IDE_ATAPI_SK_MEDIUM_ERROR      0x3
#define IDE_ATAPI_SK_HARDWARE_ERROR    0x4
#define IDE_ATAPI_SK_ILLEGAL_REQUEST   0x5
#define IDE_ATAPI_SK_UNIT_ATTENTION    0x6
#define IDE_ATAPI_SK_DATA_PROTECT      0x7
#define IDE_ATAPI_SK_ABORTED_COMMAND   0xb
#define IDE_ATAPI_SK_MISCOMPARE        0xe

// ATAPI signature
#define ATAPI_SIG_CYL_HIGH             0xeb
#define ATAPI_SIG_CYL_LOW              0x14

// Memory addresses
#define IDE1_BASE 0x1f0 // Default address for primary controller
#define IDE2_BASE 0x170 // Default address for secondary controller
#define IDE3_BASE 0x1e8 // Default address for tertiary controller
#define IDE4_BASE 0x168 // Default address for quaternary controller

#define IDE_DATA       0
#define IDE_ERROR      1
#define IDE_CYL_HIGH   4
#define IDE_CYL_LOW    5
#define IDE_DRIVE_HEAD 6
#define IDE_STATUS     7
#define IDE_ALT_STATUS 206
#define IDE_DEVCTRL    207

// Types

typedef enum _ide_dev
{
  IDE_DEV_MASTER = 0x00,
  IDE_DEV_SLAVE = 0x10
} ide_dev_t;

typedef struct
{
  char fw_rev[9];     // 8 chars + terminating 0 char
  char model_num[41]; // 40 chars + terminating 0 char
  unsigned general_config;
} ide_drive_info_t;

// Errors (see errmacro.h for explanation!)

#define FOREACH_IDE_ERROR(E) \
E(E_IDE_SUCCESS, "Error handler called on non-error.") \
E(E_IDE_TIMEOUT, "E_IDE_TIMEOUT") \
E(E_IDE_WAITBUSY, "Timeout waiting for IDE to clear busy status") \
E(E_IDE_WAITDRQ, "Timeout waiting for high DRQ") \
E(E_IDE_WAITDRDY, "Device not ready") \
E(E_IDE_IDDRIVE_NOT_ABORTED, "Identify drive command was not aborted as " \
  "expected. Please check that device is not a hard disk.") \
E(E_IDE_ATAPI_SRST_ABORTED, "ATAPI soft-reset command was aborted " \
  "unexpectedly. Maybe the device doesn't support this command.") \
E(E_IDE_NO_ATAPI_SIG,"No ATAPI signature found.") \
E(E_IDE_NOT_ATAPI_DEV, "Device does not speak ATAPI.") \
E(E_IDE_NOT_A_CDROM, "Device speaks ATAPI, but is not a CD-ROM.") \
E(E_IDE_WRONG_PKT_SIZE, "Device has wrong packet size.") \
E(E_IDE_COMMAND_ABORTED, "ATA command aborted unexpectedly.") \
E(E_IDE_UNKNOWN, "Unknown IDE/ATAPI error occured.")

typedef enum _ide_error { FOREACH_IDE_ERROR(GENERATE_ENUM_MEMBER) }
  ide_error_t;

// Prototypes

char* ide_get_error_text(ide_error_t ide_error);
unsigned ide_get_error_reg_bits();
ide_error_t ide_check_if_cdrom(ide_drive_info_t *di);
ide_error_t ide_get_atapi_info(ioport_t port, ide_dev_t dev,
  ide_drive_info_t *di);
ide_error_t ide_poll_register(ioport_t port, unsigned mask, unsigned state);
ide_error_t ide_check_error_condition(ioport_t port);
void ide_atapi_get_cd_speed_packet(unsigned speed, unsigned packet[]);
ide_error_t ide_atapi_send_packet_cmd(ioport_t port, ide_dev_t dev,
  unsigned packet[]);
ide_error_t ide_atapi_soft_reset(ioport_t port, ide_dev_t dev);

#endif
